from datetime import datetime

from sweeping_view.rmv import RMVReplay

def test_rmv(replay_path):
    rmv = RMVReplay.from_file(replay_path / "test_subject.rmv")

    assert rmv.player_data["name"] == "tkolar"
    assert rmv.properties["level"] == "beginner"
    assert rmv.get_boardgen_time() == datetime(2020, 9, 19, 21, 41, 17)
    assert set(rmv.mines) == {
        (1, 3),
        (2, 3),
        (4, 2),
        (4, 3),
        (4, 4),
        (4, 5),
        (4, 6),
        (5, 4),
        (6, 6),
        (7, 6),
    }


def test_rmv2(replay_path):
    rmv = RMVReplay.from_file(replay_path / "test_subject_2.rmv")

    assert rmv.clone_id == 1
    assert rmv.major_version_of_clone == 5
    assert rmv.format_version == 2
    assert rmv.square_size == 24
    assert rmv.player_data["name"] == "Thomas Kolar"
    assert rmv.player_data["nickname"] == "ralokt"
    assert rmv.properties["level"] == "beginner"
    assert rmv.extension_properties["vsweep_skin_fname"] == b"24px_new.bmp"
    assert set(rmv.mines) == {
        (0, 2),
        (0, 6),
        (1, 5),
        (2, 3),
        (3, 6),
        (5, 2),
        (5, 3),
        (6, 0),
        (6, 1),
        (6, 6),
    }
    assert rmv.events[-2:] == [
        {'col': 7, 'row': 7, 'subtype': 'open_1', 'type': 'board'},
        {'how': 'win', 'type': 'terminate'},
    ]



from datetime import datetime

from .base import BaseReplay
from .exceptions import InvalidReplayError


class EVFBoard:
    def __init__(self, data, cols):
        self.data = data
        self.cols = cols

    def __getitem__(self, item):
        (xx, yy) = item
        bit_index = yy * self.cols + xx
        byte_index = bit_index // 8
        return bool(self.data[byte_index] & (128 >> (bit_index % 8)))


class EVFReplay(BaseReplay):
    MODES = {
        0: "normal",
        1: "UPK",
        2: "cheat",
        3: "density",
        4: "win7",
        5: "competitive_solvable",
        6: "strong_solvable",
        7: "weak_solvable",
        8: "to_be_solvable",
        9: "strong_guessable",
        10: "weak_guessable",
        11: "chording_recursive",
        12: "flag_recursive",
        13: "chording_flag_recursive",
    }

    MOUSE_EVENT_TYPES = {
        1: "move",
        2: "lmb_down",
        3: "lmb_up",
        4: "rmb_down",
        5: "rmb_up",
        6: "mmb_down",
        7: "mmb_up",
        8: "preflag",
        9: "chord",
        10: "lmb",
        11: "rmb",
        12: "mmb",
    }

    def process_buffer(self, data):
        version_number = self.read_int(data.read(1))
        if version_number != 3:
            raise UnknownFormatVersionError(self, version_number)

        summary = self.read_int(data.read(1))
        self.completed = bool(summary & (127 >> 0))
        self.official = bool(summary & (127 >> 1))
        self.fair = bool(summary & (127 >> 2))
        nf = bool(summary & (127 >> 3))

        settings = self.read_int(data.read(1))
        qm_disabled = bool(settings & (127 >> 0))
        board_clip = bool(settings & (127 >> 1))
        loss_autorestart = bool(settings & (127 >> 2))

        self.rows = self.read_int(data.read(1))
        self.cols = self.read_int(data.read(1))
        self.num_mines = self.read_int(data.read(2))
        self.cell_size = self.read_int(data.read(1))
        game_mode_raw = self.read_int(data.read(2))
        self.bbbv = self.read_int(data.read(2))
        self.timeth = self.read_int(data.read(3))
        self.version_info = self.read_c_string(data)
        self.user_identifier = self.read_c_string(data)
        self.competition_identifier = self.read_c_string(data)
        self.unique_identifier = self.read_c_string(data)
        self.start_ts = self.read_c_string(data)
        self.end_ts = self.read_c_string(data)
        self.country_code = self.read_c_string(data)
        self.uuid = self.read_c_string(data)
        board = EVFBoard(data.read((self.cols * self.rows - 1) // 8 + 1), self.cols)

        game_mode = self.MODES.get(game_mode_raw, None)
        if game_mode is None:
            raise InvalidReplayError(
                f"Invalid game mode {game_mode_raw}!",
            )

        num_mines = 0
        self.mines = []
        for xx in range(self.cols):
            for yy in range(self.rows):
                if board[xx, yy]:
                    num_mines += 1
                    self.mines.append((yy, xx))
        if num_mines != self.num_mines:
            raise InvalidReplayError(
                "Number of mines in header field is inconsistent with the board!",
            )

        self.events = []
        while True:
            op = self.read_int(data.read(1))
            if op in (0, 255):
                break
            ts = self.read_int(data.read(3))
            xx = self.read_int(data.read(2))
            yy = self.read_int(data.read(2))
            st = self.MOUSE_EVENT_TYPES.get(op, None)
            if st is None:
                raise InvalidReplayError(f"Unknown mouse operation {op}")
            self.events.append(
                {
                    "type": "mouse",
                    "subtype": st,
                    "gametime": ts,
                    "xpos": xx,
                    "ypos": yy,
                }
            )
        if op == 0:
            self.checksum = data.read(32)
        else:
            self.checksum = None

        level = {
            (8, 8, 10): "beginner",
            (16, 16, 40): "intermediate",
            (30, 16, 40): "expert",
        }.get(
            (
                self.cols,
                self.rows,
                self.num_mines,
            ),
            "custom",
        )

        self.properties = {
            "questionmarks": not qm_disabled,
            "nonflagging": nf,
            "mode": game_mode,
            "level": level,
        }

    def read_c_string(self, data):
        return self.read_bin_c_string(data).decode("utf-8")

    def read_bin_c_string(self, data):
        result = []
        while True:
            next = data.read(1)
            if next == b"\0":
                break
            if next == b"":
                raise InvalidReplayError("EOF while reading null-terminated string")
            result.append(next)
        return b"".join(result)

    def get_best_token_source(self):
        return self.competition_identifier

    def get_boardgen_time(self):
        return datetime.fromtimestamp(int(self.start_ts) // 1000000)



# -*- coding: utf-8 -*-

from datetime import datetime
import logging

from .base import BaseReplay
from .exceptions import InvalidReplayError, UnknownFormatVersionError

logger = logging.getLogger(__name__)


class RMVReplay(BaseReplay):
    MODES = {
        0: "normal",
        1: "UPK",
        2: "cheat",
        3: "density",
    }

    LEVELS = {
        0: "beginner",
        1: "intermediate",
        2: "expert",
        3: "custom",
    }

    MOUSE_EVENT_TYPES = {
        1: "move",
        2: "lmb_down",
        3: "lmb_up",
        4: "rmb_down",
        5: "rmb_up",
        6: "mmb_down",
        7: "mmb_up",
        28: "move",  # reduced version
    }

    BOARD_EVENT_TYPES = {
        9: "pressed",
        10: "pressed_qm",
        11: "closed",
        12: "qm",
        13: "flag",
        14: "open",
        18: "open_0",
        19: "open_1",
        20: "open_2",
        21: "open_3",
        22: "open_4",
        23: "open_5",
        24: "open_6",
        25: "open_7",
        26: "open_8",
        27: "open_blast",
    }

    TERMINATION_EVENT_TYPES = {
        15: "blast",
        16: "win",
        17: "other",
    }

    PLAYER_FIELDS = (
        "name",
        "nickname",
        "country",
        "token",
    )

    def process_buffer(self, data):
        # header 1
        extension = data.read(4)
        self.format_version = self.read_int(data.read(2))

        if not (1 <= self.format_version <= 2):
            raise UnknownFormatVersionError(self, self.format_version)

        self.clone_id = None
        self.major_version_of_clone = None
        if self.format_version >= 2:
            self.clone_id = self.read_int(data.read(1))
            self.major_version_of_clone = self.read_int(data.read(1))

        filesize = self.read_int(data.read(4))

        # header 2
        if self.format_version == 1:
            result_str_size = self.read_int(data.read(2))
        version_info_size = self.read_int(data.read(2))
        player_info_size = self.read_int(data.read(2))
        board_size = self.read_int(data.read(2))
        preflagged_size = self.read_int(data.read(2))
        properties_size = self.read_int(data.read(2))
        if self.format_version >= 2:
            extension_properties_size = self.read_int(data.read(2))
        vid_size = self.read_int(data.read(4))
        checksum_size = self.read_int(data.read(2))

        if self.format_version == 1:
            # result string
            result_str = data.read(result_str_size)

        # version information
        self.version_info = data.read(version_info_size)

        # player fields
        num_player_fields = self.read_int(data.read(2))
        player_fields = []
        player_data = {}
        for _ in range(num_player_fields):
            field_size = ord(data.read(1))
            player_fields.append(data.read(field_size))

        for field_name, field_value in zip(self.PLAYER_FIELDS, player_fields):
            player_data[field_name] = field_value

        # board
        self.timestamp_boardgen = self.read_int(data.read(4))
        self.cols = ord(data.read(1))
        self.rows = ord(data.read(1))
        self.num_mines = self.read_int(data.read(2))
        self.mines = []
        for _ in range(self.num_mines):
            col = ord(data.read(1))
            row = ord(data.read(1))
            self.mines.append((row, col))

        # preflagged
        self.preflags = []
        if preflagged_size:
            num_preflags = self.read_int(data.read(2))
            for _ in range(num_preflags):
                col = ord(data.read(1))
                row = ord(data.read(1))
                self.preflags.append((row, col))

        # properties
        properties = []
        self.properties = {}
        for _ in range(properties_size):
            properties.append(ord(data.read(1)))

        utf8 = self.format_version >= 2
        self.square_size = 16
        bbbv = None

        self.properties["questionmarks"] = bool(properties[0])
        self.properties["nonflagging"] = bool(properties[1])
        try:
            self.properties["mode"] = self.MODES[properties[2]]
        except KeyError:
            raise InvalidReplayError(self, message="Invalid mode!")
        try:
            self.properties["level"] = self.LEVELS[properties[3]]
        except KeyError:
            raise InvalidReplayError(self, message="Invalid level!")
        if self.format_version == 1 and properties_size > 4:
            utf8 = properties[4]
        if self.format_version >= 2:
            bbbv = properties[4] + (properties[5] << 8)
            self.square_size = properties[6]

        # TODO: verify that cp1252 is correct, or add encoding parameter
        encoding = "utf-8" if utf8 else "cp1252"

        self.player_data = {
            name: value.decode(encoding) for name, value in player_data.items()
        }

        result_str_field_list = (
            [i.strip() for i in result_str.decode(encoding).split("#") if i.strip()]
            if self.format_version == 1
            else []
        )
        self.result_str_dict = {
            key.strip(): value.strip()
            for key, value in (field.split(":") for field in result_str_field_list)
        }
        self.bbbv = int(self.result_str_dict["3BV"]) if bbbv is None else bbbv

        self.extension_properties = {}
        if self.format_version >= 2:
            num_properties = self.read_int(data.read(2))
            for _ in range(num_properties):
                key_size = ord(data.read(1))
                key = data.read(key_size).decode(encoding)
                value_size = ord(data.read(1))
                value = data.read(value_size)
                self.extension_properties[key] = value

        self.events = []
        xpos = None
        ypos = None
        gametime = None
        nFlags = None
        (xoffs, yoffs) = (12, 56) if self.format_version == 1 else (0, 0)
        while data:
            evcode = ord(data.read(1))
            if evcode == 0:
                logger.warning("Warning, timestampchange is deprecated!")
                new_timestamp = self.read_int(data.read(4))
                self.events.append(
                    {
                        "type": "timestamp_change",
                        "new_timestamp": new_timestamp,
                    }
                )
            elif 1 <= evcode <= 7 or evcode == 28:
                if evcode == 28:
                    if gametime is None or xpos is None or ypos is None:
                        raise InvalidReplayError(
                            self,
                            message="first mouse event was reduced mouse move",
                        )
                    gametime += ord(data.read(1))
                    mv = ord(data.read(1))
                    # two 4bit two's complement signed integers
                    # n & 7 = last three digits
                    # n & 8 = the leading digit (that has a negative weight in
                    # two's complement)
                    # and of course the xpos change gets shifted into place
                    xpos += (mv >> 4) & 7
                    xpos -= (mv >> 4) & 8
                    ypos += mv & 7
                    ypos -= mv & 8
                else:
                    gametime = self.read_int(data.read(3))
                    nFlags = ord(data.read(1))
                    xpos = self.read_int(data.read(2))
                    ypos = self.read_int(data.read(2))
                self.events.append(
                    {
                        "type": "mouse",
                        "subtype": self.MOUSE_EVENT_TYPES[evcode],
                        "gametime": gametime,
                        "nFlags": nFlags,
                        # in RMV v1, these coordinates are relative to the top
                        # right corner of the client area (ie, the whole UI,
                        # including borders and top bar)
                        # in later versions, they are relative to the top left
                        # corner of the board
                        "xpos": xpos - xoffs,
                        "ypos": ypos - yoffs,
                    }
                )

            elif 9 <= evcode <= 14 or 18 <= evcode <= 27:
                col = ord(data.read(1))
                row = ord(data.read(1))
                self.events.append(
                    {
                        "type": "board",
                        "subtype": self.BOARD_EVENT_TYPES[evcode],
                        "col": col,
                        "row": row,
                    }
                )

            elif 15 <= evcode <= 17:
                self.events.append(
                    {
                        "type": "terminate",
                        "how": self.TERMINATION_EVENT_TYPES[evcode],
                    }
                )
                break
            else:
                raise InvalidReplayError(self)

        self.timeth = self.read_int(data.read(3))
        self.checksum = data.read(checksum_size)

    def get_best_token_source(self):
        token = self.player_data.get("token", None)
        if token is None:
            token = self.result_str_dict["NICK"]
        return token

    def get_boardgen_time(self):
        return datetime.fromtimestamp(self.timestamp_boardgen)
